// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.
/***/
/*
It assigns for each input vertex/point a corresponding LIDAR vertex to calculate point-wise metrics downstream in the pipeline
Outputs:
- indexpoint: buffer with the ID correspondences between 2D grid points of input solution and Lidar data.
- valids: buffer with the validity flag equal to 1 to indicate that a point withing a pointspacing distance has been assigned, 0 otherwise

Inputs:
- x2: buffer containing the X-coordinate 2D grid values of the input solution data
- y2: buffer containing the Y-coordinate 2D grid values of the input solution data
- z2: buffer containing the Z-coordinate 2D grid values of the input solution data
- x1: buffer containing the X-coordinate 2D grid values of the Lidar data
- y1: buffer containing the Y-coordinate 2D grid values of the Lidar data
- z1: buffer containing the Z-coordinate 2D grid values of the Lidar data
- radoffset: search radius in the 2-D grid to find potential points within pointspacing distance. (This is to avoid searching over millions of points everytime).
- rowIN: number of rows in the 2D input grid
- colIN: number of cols in the 2D input grid
- pointspacing: this is the distance separation between cells in the 2D grid.

Version history: 2017-03-28	Marc Bosch	Initial version
*/
/***/
#include <math.h>
#include "mex.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits>

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{	
    if(nlhs > 2){
        mexErrMsgTxt("Function has two return value");
        return;
    }

    if(nrhs !=10){
        mexErrMsgTxt("Usage: mexfindpointreferenceC(double *X2, double *Y2, double *Z2, double *X, double *Y, double *Z, double *radpffset,double *rowIN,double *colIN, double *pointspacing)");
        return;
    }
    
    /* get the image buffer*/    
    double *X2 = (double*) mxGetPr(prhs[0]);
	double *Y2 = (double*) mxGetPr(prhs[1]);
	double *Z2 = (double*) mxGetPr(prhs[2]);
    int* dims = (int*) mxGetDimensions(prhs[0]);
    int rows = dims[1];//flipped since it comes transpose and row major
    int cols = dims[0];
	double *Xref = (double*) mxGetPr(prhs[3]);
	double *Yref = (double*) mxGetPr(prhs[4]);
	double *Zref = (double*) mxGetPr(prhs[5]);
	double* pradoffset = mxGetPr(prhs[6]);
 	double radoffset = *pradoffset;
	double* prowIN = mxGetPr(prhs[7]);
 	double rowIN = *prowIN;
	double* pcolIN = mxGetPr(prhs[8]);
 	double colIN = *pcolIN;
	double* ppointspacing = mxGetPr(prhs[9]);
 	double pointspacing = *ppointspacing;
	
    int* dimsbuf = (int*) mxGetDimensions(prhs[3]);
 	int len = dimsbuf[0]*dimsbuf[1];
    plhs[0] = mxCreateDoubleMatrix(rows,cols, mxREAL);
    double *indexpoint = mxGetPr(plhs[0]);  
    for (int i=0; i < rows*cols; i++)
        indexpoint[i] = 0;//dims[0];  
	plhs[1] = mxCreateDoubleMatrix(rows,cols, mxREAL);
    double *valids = mxGetPr(plhs[1]);  
    for (int i=0; i < rows*cols; i++)
        valids[i] = 0;//dims[0];  
		
   //look for closer geometric point and assign the corresponding z value and record the point index
	int kistart=0;
	int kiend=dimsbuf[1];
	int kjstart=0;
	int kjend=dimsbuf[0];
	int kibest=0,kjbest=0;
	int kistart_last=0;
	int kjstart_last=0;
	int searchoffset=radoffset;
	for (int i=0;i<rows*cols;i++)
	{
		double dmin = std::numeric_limits<double>::max();
		for(int ki=kistart;ki<kiend;ki++)
			for(int kj=kjstart;kj<kjend;kj++)
			{
				int k=ki*dimsbuf[0]+kj;
				double dist=((X2[i]-Xref[k])*(X2[i]-Xref[k])+
					(Y2[i]-Yref[k])*(Y2[i]-Yref[k]));/*+
					(Z2[i]-Zref[k])*(Z2[i]-Zref[k]));*/
				if (dist<dmin)
				{
					dmin=dist;
					indexpoint[i] = (double)(k+1);//matlab index starts at 1
					kibest=ki;
					kjbest=kj;
					
				}

			}
		if (dmin<pointspacing*pointspacing*2)
			valids[i]=1;
		//this allows to start the search around the best coordinate from the previous iteration (speed up) and avoid full search
		kistart=MAX(0,kibest-searchoffset);
		kiend=MIN(dimsbuf[1],kibest+searchoffset);
		kjstart=MAX(0,kjbest-searchoffset);
		kjend=MIN(dimsbuf[0],kjbest+searchoffset);
		//record best location for each new column/row
		if(i%(int)colIN==0)
		{
			kistart_last = kibest;
			kjstart_last = kjbest;
		}
		//re-set for each new column/row
		if(i%(int)colIN==(int)colIN-1)
		{
			//kistart=0;
			//kiend=dimsbuf[1];
			//kjstart=0;
			//kjend=dimsbuf[0];
			kistart=MAX(0,kistart_last-searchoffset);
			kiend=MIN(dimsbuf[1],kistart_last+searchoffset);
			kjstart=MAX(0,kjstart_last-searchoffset);
			kjend=MIN(dimsbuf[0],kjstart_last+searchoffset);			
		}

	}




	//for (int i=0;i<rows*cols;i++)
	//{
	//	double dmin = std::numeric_limits<double>::max();
	//	for (int k=0;k<len;k++)
	//	{
	//		double dist=((X2[i]-Xref[k])*(X2[i]-Xref[k])+
	//		(Y2[i]-Yref[k])*(Y2[i]-Yref[k])+
	//		(Z2[i]-Zref[k])*(Z2[i]-Zref[k]));
	//		if (dist<dmin)
	//		{
	//			dmin = dist;
	//			indexpoint[i] = (double)(k+1);//matlab index starts at 1

	//		}
	//	}
	//}


	//assign for each ground truth point the closest input point
	//for(int k=0;k<len;k++)
	//{

	//	double dmin = std::numeric_limits<double>::max();
	//	for (int i=0;i<rows*cols;i++)
	//	{
	//		double dist=((X2[i]-Xref[k])*(X2[i]-Xref[k])+
	//		(Y2[i]-Yref[k])*(Y2[i]-Yref[k])+
	//		(Z2[i]-Zref[k])*(Z2[i]-Zref[k]));
	//		//replace if it equal or smaller
	//		if(dist<dmin)
	//		{
	//			dmin=dist;
	//			xinputbuf[k] = X2[i];
	//			yinputbuf[k] = Y2[i];
	//			zinputbuf[k] = Z2[i];			
	//		}	
	//	}

	//}

	
   
   
   
}
